#!/usr/bin/python

"""
netsim

A network simulator platform for validating distributed system implementations.

Basic idea:

* Each node is represented by a deterministic state machine.
* The state machine is changed by delivering events.
* A sequence of events leads a machine to a unique state.

Thus, the log of all events arriving at a state machine summarizes its state.
The state of the entire system is represented by the set of logs for all state
machines. Recording the order that we choose to deliver messages allows us to
replay the simulation to the exact same state.

We reduce the search space by not visiting states that have the same sets of
logs. We summarize each state machine's log using a cryptographic hash. We then
combine each state machine's "history hash" using the hash function. This means
we have a very low probability of collisions, which would result in us not
visiting some states that need to be visited.

To generate a unique id for messages we hash their serialized representation.
We used to use the source's hash ^ message_round_count. Using the serialized
representation means that if for some reason a state machine sends the same
response to the sequence (a, b, c) as it does to the sequence (b, a, c), then
we may be able to prune the search space. TODO: Is this actually possible?
for the small Paxos examples it is not.

TODO: Hashing message contents is MUCH slower than the source id approach.

TODO: pickle the node state and use that to represent the state? (+order independent hash?)
TODO: support replicas using an order-independent hash function
TODO: represent the choices as a string rather than a list?


The visit order changes how many executions need to be performed. The default
FIFO order seems to work the best. Doing a breadth first traversal is the worst,
and LIFO order is also bad.
"""

# TODO: Add support for replica groups?
# TODO: Instead of deleting from the middle of the event_queue, swap with the last element?

import hashlib
import struct


HASH_TYPE = hashlib.sha1


class Message(object):
    """Contains data along with header information."""

    __slots__ = ("source", "destination", "id", "data")

    def __init__(self, id, source, destination, data):
        assert source >= 0
        assert destination >= 0
        assert source != destination

        self.source = source
        self.destination = destination
        self.id = id
        self.data = data

    def __eq__(self, other):
        equal = self.id == other.id
        # Don't assert data is equal: it might be unserialized messages that don't implement __eq__
        assert (not equal) or (self.source == other.source and self.destination == other.destination)
        return equal

    def __ne__(self, other):
        return not self.__eq__(other)


class NodeState(object):
    """Encapsulates the current state of a node.
    
    The history of events is represented by hashing IDs. Each time a message is
    received, the id of the message is hashed with the current history hash.
    Before a message is sent, the hishory ID is hashed with itself. The
    message's id is equal to the state that generated it.
    """

    __slots__ = ("state_machine", "history_hash", "hash_cache", "message_round_count")

    def __init__(self, state_machine, initial_id):
        """initial_id must be unique so node history ids are unique."""

        assert state_machine is not None
        assert initial_id >= 0
        #~ assert len(initial_id) == HASH_LENGTH
        self.state_machine = state_machine
        self.history_hash = HASH_TYPE(struct.pack("L", initial_id))
        self.hash_cache = self.history_hash.digest()
        self.message_round_count = 0

    def historyHash(self):
        """Returns the hash of this node's sequence of operations."""
        return self.hash_cache

    def messageReceived(self, system, message):
        """Deliver message to the state machine in this node."""

        # Update the history hash then deliver the message
        self.history_hash.update(message.id)
        self.hash_cache = self.history_hash.digest()
        self.message_round_count = 0
        self.state_machine.messageReceived(system, message.source, message.data)

    def getSentMessageId(self):
        """Returns the id of a message to be sent."""
        message_id = self.historyHash()
        
        # Message id is the node's state XOR sent count
        if self.message_round_count > 0:
            last_byte = ord(message_id[-1])
            assert self.message_round_count < 256
            last_byte ^= self.message_round_count
            message_id = message_id[:-1] + chr(last_byte)
        self.message_round_count += 1

        return message_id


class VisitedStates(object):
    __slots__ = ("delivered", "more_events")

    def __init__(self, num_nodes):
        self.delivered = [set() for _ in xrange(num_nodes)]
        self.more_events = False

    def can_deliver(self, destination, event_id):
        return event_id not in self.delivered[destination]

    def deliver(self, destination, event_id):
        """Returns a new VisitedStates object that represents the state after delivering an event
        to destination. If we are not permitted to deliver event_id, returns None."""

        assert self.can_deliver(destination, event_id)
        self.delivered[destination].add(event_id)

        # Clone this object
        result = VisitedStates(len(self.delivered))
        for i, s in enumerate(self.delivered):
            result.delivered[i] = set(s)

        # Replace the destination with the empty set as all messages are now
        # permitted
        result.delivered[destination] = set()

        return result


class System(object):
    """Provides a framework for simulating asynchronous networks."""

    __slots__ = ("nodes", "event_queue", "current_node")

    def __init__(self):
        self.nodes = []
        # Holds pending events for the current simulation
        self.event_queue = []
        self.current_node = None

    def addNode(self, node):
        assert node not in [node_state.state_machine for node_state in self.nodes]
        node_id = len(self.nodes)
        self.nodes.append(NodeState(node, node_id))
        return node_id

    def startStateMachines(self):
        """Initializes all nodes by calling their start methods."""
        assert len(self.nodes) > 0
        assert len(self.event_queue) == 0
        for node_id, node in enumerate(self.nodes):
            self.current_node = node_id
            node.state_machine.start(self, node_id)
            self.current_node = None

    def executeChoice(self, choice):
        # Remove the event from the queue
        event = self.event_queue[choice]
        del self.event_queue[choice]
        self.current_node = event.destination
        #~ print choice, event.source, "->", event.destination, event.data, self.current_node
        self.nodes[self.current_node].messageReceived(self, event)
        #~ def summary(hash):
            #~ bytes = hash.digest()[:4]
            #~ return "".join(["%02x" % ord(x) for x in bytes])
        #~ print [summary(n.history_hash) for n in self.nodes]
        self.current_node = None

    def simulateDeterministic(self):
        assert len(self.nodes) > 0
        assert len(self.event_queue) == 0

        # Simulate until there are no more events
        self.startStateMachines()
        while len(self.event_queue) > 0:
            self.executeChoice(0)

    def resimulate(self, initializer, choices):
        """Simulates the events represented by choices."""
        # Call the initializer to create the initial set of nodes
        # TODO: Test determinism?
        self.nodes = []
        self.event_queue = []
        initializer(self)
        self.startStateMachines()

        # Replay history
        for choice in choices:
            self.executeChoice(choice)
    
    def sendMessage(self, destination, message):
        assert self.current_node is not None
        assert destination is not None
        assert destination < len(self.nodes)

        message_id = self.nodes[self.current_node].getSentMessageId()
        self.event_queue.append(Message(message_id, self.current_node, destination, message))


def listIsPrefix(l, prefix):
    """Returns true if prefix is a prefix for l."""

    if len(l) < len(prefix):
        return False

    # Check using Python slice notation
    return l[:len(prefix)] == prefix


class SimulatedSystem(object):
    __slots__ = (
            "initializer", "initial_states", "initial_events", "states", "state_transitions",
            "choices", "saved_states", "saved_events", "event_queue", "system", "system_choices")

    def __init__(self, initializer):
        # Initialize a system to get the number of nodes
        self.initializer = initializer
        self.resetSystem()

        self.initial_states = [n.historyHash() for n in self.system.nodes]
        self.initial_events = list(self.system.event_queue)
        self.state_transitions = [{} for _ in xrange(len(self.initial_states))]

        self.reset()

    def numNodes(self):
        return len(self.initial_states)

    def reset(self):
        """Resets the simulated system state."""
        self.choices = []
        self.saved_states = [tuple(self.initial_states)]
        self.saved_events = [tuple(self.initial_events)]
        self.event_queue = tuple(self.initial_events)
        self.states = tuple(self.initial_states)

    def resetSystem(self):
        self.system = System()
        self.system_choices = []
        self.initializer(self.system)
        self.system.startStateMachines()

    def addTransition(self, node_id, state, message, out_state, new_messages):
        key = (state, message)
        assert key not in self.state_transitions[node_id]
        self.state_transitions[node_id][key] = (out_state, new_messages)

    def getNextStateMessages(self, node_id, state, message_id):
        key = (state, message_id)
        return self.state_transitions[node_id].get(key, (None, None))

    def resimulate(self, choices):
        # Find the divergence point
        i = 0
        truncate = min(len(choices), len(self.choices))
        while i < truncate:
            if self.choices[i] != choices[i]:
                break
            i += 1

        # Truncate choices at i
        del self.choices[i:]
        #~ assert listIsPrefix(choices, self.choices)

        # Restore state at i
        del self.saved_states[i+1:]
        self.states = list(self.saved_states[i])
        del self.saved_events[i+1:]
        self.event_queue = list(self.saved_events[i])

        # Execute forward
        for choice in choices[i:]:
            self.executeChoice(choice)
        #~ assert self.choices == choices

    def executeChoice(self, choice):
        # Check for cached state
        event = self.event_queue[choice]
        current_state = self.states[event.destination]
        next_state, new_messages = self.getNextStateMessages(
                event.destination, current_state, event.id)

        executed = False
        if next_state is None:
            self.simulateUpToDate()

            # Execute the new choice, capturing state
            new_event_index = len(self.system.event_queue) - 1
            self.system_choices.append(choice)
            self.system.executeChoice(choice)
            next_state = self.system.nodes[event.destination].historyHash()
            new_messages = self.system.event_queue[new_event_index:]

            self.addTransition(event.destination, current_state, event.id, next_state, new_messages)
            executed = True

        # "Execute" the choice
        del self.event_queue[choice]
        self.choices.append(choice)
        self.states[event.destination] = next_state
        self.event_queue.extend(new_messages)
        self.saved_states.append(tuple(self.states))
        self.saved_events.append(tuple(self.event_queue))

        if executed:
            assert self.event_queue == self.system.event_queue
            assert self.states == [n.historyHash() for n in self.system.nodes]

    def simulateUpToDate(self):
        """Bring the real simulation up to date with the fake simulation."""
        # Need to *actually* execute the choice: determine if it is a prefix
        if not listIsPrefix(self.choices, self.system_choices):
            # It is NOT a prefix: reset the system
            self.resetSystem()

        # Execute the missing tail
        assert listIsPrefix(self.choices, self.system_choices)
        for c in self.choices[len(self.system_choices):]:
            self.system_choices.append(c)
            self.system.executeChoice(c)

        assert self.system_choices == self.choices
        assert len(self.system.event_queue) == len(self.event_queue)
        assert self.system.event_queue == self.event_queue

    def stateMachines(self):
        self.simulateUpToDate()
        return [n.state_machine for n in self.system.nodes]


def simulateAll(initializer, validator):
    """Simulates all possible outcomes.
    
    initializer: called to create the initial state in a deterministic fashion
    validator: returns true if the final state is okay."""

    system = SimulatedSystem(initializer)

    # sequences of choices that fail validation
    error_choices = []

    choices = []
    history_stack = [VisitedStates(system.numNodes())]

    execute_count = 0
    reject_count = 0
    while len(history_stack) > 0:
        execute_count += 1

        # Replay history: we end up at a new state
        system.resimulate(choices)
        #~ print choices

        # Make new choices
        while len(system.event_queue) > 0:
            # Look for a new choice
            history = history_stack[-1]
            
            permitted_choice = None
            history.more_events = False  # Nothing more to visit?
            for choice, event in enumerate(system.event_queue):
                if history.can_deliver(event.destination, event.id):
                    if permitted_choice is None:
                        # We found one: we are going to visit it now so record the history
                        permitted_choice = choice
                        new_history = history.deliver(event.destination, event.id)
                        history_stack.append(new_history)
                    else:
                        # There is at least one more choice to revisit here
                        assert history.more_events is False
                        history.more_events = True
                        break
                else:
                    reject_count += 1
            
            if permitted_choice is None:
                # Nothing remaning to visit: stop execution
                break

            # execute the choice
            choices.append(permitted_choice)
            system.executeChoice(permitted_choice)

        # This if statement is false if we aborted due to duplicate history
        if len(system.event_queue) == 0:
            # We have reached the end of execution: validate
            state_machines = system.stateMachines()
            if not validator(state_machines):
                print "error:", choices
                error_choices.append(choices)

        # Pop off choices until we have more choices to make
        while len(history_stack) > 0 and not history_stack[-1].more_events:
            if len(choices) == 0:
                choices = None
            else:
                choices.pop()
            assert not history_stack[-1].more_events
            history_stack.pop()
        assert len(history_stack) == 0 or history_stack[-1].more_events

    # We have exhausted the universe of possible orderings
    print execute_count, reject_count
    return error_choices
