#!/usr/bin/python

__pychecker__ = 'unusednames=node_id'

import unittest

import netsim


class MessageTest(unittest.TestCase):
    def testEquals(self):
        m1a = netsim.Message(0, 1, 2, "foo")
        badMessage = netsim.Message(0, 1, 3, "foo")  # same id, different contents = not permitted
        m2 = netsim.Message(1, 1, 3, "foo")
        m1b = netsim.Message(0, 1, 2, "foo")

        self.assertRaises(AssertionError, m1a.__eq__, badMessage)

        self.assertEquals(m1a, m1b)
        self.assertNotEquals(m1a, m2)
        self.assertNotEquals(m1b, m2)


class MockStateMachine(object):
    def __init__(self, start_destination):
        self._start = start_destination

    def messageReceived(self, system, source, payload):
        self.system = system
        self.source = source
        self.payload = payload

    def start(self, system, node_id):
        if self._start is not None:
            system.sendMessage(self._start, "foo")


class MockSystem(object):
    pass


class NodeStateTest(unittest.TestCase):
    def setUp(self):
        self.system = MockSystem()
        self.machine = MockStateMachine(None)
        self.state = netsim.NodeState(self.machine, 0)
        self.initial_history = self.state.historyHash()

    def testBadConstructor(self):
        self.assertRaises(AssertionError, netsim.NodeState, None, 0)
        self.assertRaises(AssertionError, netsim.NodeState, self.machine, None)

    def testReceiveUpdatesState(self):
        message = netsim.Message("id", 0, 1, "data")
        self.state.messageReceived(self.system, message)
        self.assertEquals(self.system, self.machine.system)
        self.assertEquals(0, self.machine.source)
        self.assertEquals("data", self.machine.payload)
        self.assertNotEquals(self.initial_history, self.state.historyHash())

    def testSendMessageUpdatesState(self):
        message_id = self.state.getSentMessageId()
        message_id2 = self.state.getSentMessageId()
        self.assertNotEquals(message_id, message_id2)

        # Reset the state: we should generate the same ids again
        self.state = netsim.NodeState(self.machine, 0)
        self.assertEquals(message_id, self.state.getSentMessageId())
        self.assertEquals(message_id2, self.state.getSentMessageId())


class VisitedStateTest(unittest.TestCase):
    def testSimple(self):
        visited = netsim.VisitedStates(2)
        self.assertTrue(visited.can_deliver(0, "foo"))
        visited2 = visited.deliver(0, "foo")
        self.assertFalse(visited.can_deliver(0, "foo"))
        self.assertTrue(visited2.can_deliver(0, "foo"))

        self.assertTrue(visited.can_deliver(1, "foo"))
        visited2 = visited.deliver(1, "foo")
        self.assertFalse(visited.can_deliver(1, "foo"))
        self.assertTrue(visited2.can_deliver(1, "foo"))

        self.assertFalse(visited.can_deliver(0, "foo"))
        self.assertTrue(visited.can_deliver(0, "bar"))
        visited2 = visited.deliver(0, "bar")
        self.assertFalse(visited.can_deliver(0, "bar"))
        self.assertTrue(visited2.can_deliver(0, "bar"))
        self.assertTrue(visited2.can_deliver(0, "foo"))


def createNetwork(system):
    receiver = MockStateMachine(None)
    destination = system.addNode(receiver)
    
    sender = MockStateMachine(destination)
    system.addNode(sender)


def validateFail(nodes):
    for node in nodes:
        if node._start is None:
            assert node.payload == "foo"
    
    return False


def validateSuccess(nodes):
    return True


class SystemTest(unittest.TestCase):
    def setUp(self):
        self.system = netsim.System()
        self.machine = MockStateMachine(None)

    def testAdd(self):
        self.system.addNode(self.machine)
        self.assertRaises(AssertionError, self.system.addNode, self.machine)
        self.system.addNode(MockStateMachine(None))

class SimulateTest(unittest.TestCase):
    def testSimpleError(self):
        errors = netsim.simulateAll(createNetwork, validateFail)
        self.assertEquals([[]], errors)
        errors = netsim.simulateAll(createNetwork, validateSuccess)
        self.assertEquals([], errors)


class ListIsPrefix(unittest.TestCase):
    def testEmpty(self):
        self.assertTrue(netsim.listIsPrefix([], []))

    def testSingle(self):
        self.assertTrue(netsim.listIsPrefix([1], []))
        self.assertTrue(netsim.listIsPrefix([1], [1]))
        self.assertFalse(netsim.listIsPrefix([1], [2]))

    def testPrefix(self):
        self.assertTrue(netsim.listIsPrefix([1, 2, 3], []))
        self.assertTrue(netsim.listIsPrefix([1, 2, 3], [1]))
        self.assertTrue(netsim.listIsPrefix([1, 2, 3], [1, 2]))
        self.assertTrue(netsim.listIsPrefix([1, 2, 3], [1, 2, 3]))

    def testNotPrefix(self):
        self.assertFalse(netsim.listIsPrefix([1, 2, 3], [2]))
        self.assertFalse(netsim.listIsPrefix([1, 2, 3], [1, 2, 3, 4]))
        self.assertFalse(netsim.listIsPrefix([1, 2, 3], [2, 2, 3]))
        self.assertFalse(netsim.listIsPrefix([1, 2, 3], [1, 3, 3]))
        self.assertFalse(netsim.listIsPrefix([1, 2, 3], [1, 2, 4]))
        self.assertFalse(netsim.listIsPrefix([1, 2, 3], [3, 2]))


if __name__ == "__main__":
    unittest.main()
